from __future__ import annotations

import asyncio
import inspect
import json
from pathlib import Path
import queue
import threading
from collections.abc import Callable
from typing import Any
from uuid import uuid4

from chainlit.data import get_data_layer
from chainlit.types import Pagination, ThreadFilter
from chainlit.user import User

from .errors import AppClosedError, DataPersistenceNotEnabledError
from .models import IncomingMessage, OutgoingCommand
from .runtime import get_runtime


class EasierlitApp:
    """
    Communication bridge between Chainlit callbacks and user run_func.

    Easierlit runs with a single thread worker. This object is the primary API
    surface for worker message and thread CRUD operations.
    """

    _THOUGHT_TOOL_NAME = "Reasoning"
    _MESSAGE_STEP_TYPES = frozenset(
        {
            "user_message",
            "assistant_message",
            "system_message",
            "tool",
        }
    )

    def __init__(
        self,
        *,
        runtime=None,
        data_layer_getter: Callable[[], Any | None] = get_data_layer,
        uuid_factory: Callable[[], Any] = uuid4,
    ):
        self._incoming_queue: queue.Queue[IncomingMessage | None] = queue.Queue()
        self._outgoing_queue: queue.Queue[OutgoingCommand] = queue.Queue()
        self._closed = threading.Event()
        self._runtime = runtime if runtime is not None else get_runtime()
        self._data_layer_getter = data_layer_getter
        self._uuid_factory = uuid_factory

    def recv(self, timeout: float | None = None) -> IncomingMessage:
        if self._closed.is_set():
            raise AppClosedError("EasierlitApp is closed.")

        try:
            if timeout is None:
                item = self._incoming_queue.get()
            else:
                item = self._incoming_queue.get(timeout=timeout)
        except queue.Empty as exc:
            raise TimeoutError("Timed out waiting for an incoming message.") from exc

        if item is None:
            raise AppClosedError("EasierlitApp is closed.")

        return item

    async def arecv(self, timeout: float | None = None) -> IncomingMessage:
        return await asyncio.to_thread(self.recv, timeout)

    def enqueue(
        self,
        thread_id: str,
        content: str,
        *,
        session_id: str = "external",
        author: str = "User",
        message_id: str | None = None,
        metadata: dict[str, Any] | None = None,
        elements: list[Any] | None = None,
        created_at: str | None = None,
    ) -> str:
        if not isinstance(thread_id, str) or not thread_id.strip():
            raise ValueError("thread_id must be a non-empty string.")
        if not isinstance(session_id, str) or not session_id.strip():
            raise ValueError("session_id must be a non-empty string.")
        if not isinstance(author, str) or not author.strip():
            raise ValueError("author must be a non-empty string.")

        resolved_message_id = message_id
        if resolved_message_id is None:
            resolved_message_id = self._new_message_id()
        elif not isinstance(resolved_message_id, str) or not resolved_message_id.strip():
            raise ValueError("message_id must be a non-empty string when provided.")

        resolved_elements = elements or []
        resolved_metadata = metadata or {}

        self._enqueue_outgoing_command(
            command="add_message",
            thread_id=thread_id,
            message_id=resolved_message_id,
            content=content,
            author=author,
            step_type="user_message",
            metadata=resolved_metadata,
            elements=resolved_elements,
        )

        incoming = IncomingMessage(
            thread_id=thread_id,
            session_id=session_id,
            message_id=resolved_message_id,
            content=content,
            elements=resolved_elements,
            author=author,
            created_at=created_at,
            metadata=resolved_metadata,
        )
        self._enqueue_incoming(incoming)
        return resolved_message_id

    def add_message(
        self,
        thread_id: str,
        content: str,
        author: str = "Assistant",
        metadata: dict | None = None,
        elements: list[Any] | None = None,
    ) -> str:
        """Enqueue an assistant message and return the generated message id."""
        message_id = self._new_message_id()
        self._enqueue_outgoing_command(
            command="add_message",
            thread_id=thread_id,
            message_id=message_id,
            content=content,
            author=author,
            metadata=metadata,
            elements=elements,
        )
        return message_id

    def add_tool(
        self,
        thread_id: str,
        tool_name: str,
        content: str,
        metadata: dict | None = None,
        elements: list[Any] | None = None,
    ) -> str:
        """Enqueue a tool-call step and return the generated message id."""
        message_id = self._new_message_id()
        self._enqueue_outgoing_command(
            command="add_tool",
            thread_id=thread_id,
            message_id=message_id,
            content=content,
            author=tool_name,
            metadata=metadata,
            elements=elements,
        )
        return message_id

    def add_thought(
        self,
        thread_id: str,
        content: str,
        metadata: dict | None = None,
        elements: list[Any] | None = None,
    ) -> str:
        """Enqueue a reasoning step as a tool-call step."""
        return self.add_tool(
            thread_id=thread_id,
            tool_name=self._THOUGHT_TOOL_NAME,
            content=content,
            metadata=metadata,
            elements=elements,
        )

    def update_message(
        self,
        thread_id: str,
        message_id: str,
        content: str,
        metadata: dict | None = None,
        elements: list[Any] | None = None,
    ) -> None:
        self._enqueue_outgoing_command(
            command="update_message",
            thread_id=thread_id,
            message_id=message_id,
            content=content,
            metadata=metadata,
            elements=elements,
        )

    def update_tool(
        self,
        thread_id: str,
        message_id: str,
        tool_name: str,
        content: str,
        metadata: dict | None = None,
        elements: list[Any] | None = None,
    ) -> None:
        self._enqueue_outgoing_command(
            command="update_tool",
            thread_id=thread_id,
            message_id=message_id,
            content=content,
            author=tool_name,
            metadata=metadata,
            elements=elements,
        )

    def update_thought(
        self,
        thread_id: str,
        message_id: str,
        content: str,
        metadata: dict | None = None,
        elements: list[Any] | None = None,
    ) -> None:
        self.update_tool(
            thread_id=thread_id,
            message_id=message_id,
            tool_name=self._THOUGHT_TOOL_NAME,
            content=content,
            metadata=metadata,
            elements=elements,
        )

    def delete_message(self, thread_id: str, message_id: str) -> None:
        self._enqueue_outgoing_command(
            command="delete",
            thread_id=thread_id,
            message_id=message_id,
        )

    def list_threads(
        self,
        first: int = 20,
        cursor: str | None = None,
        search: str | None = None,
        user_identifier: str | None = None,
    ):
        data_layer = self._get_data_layer_or_raise()

        async def _list_threads():
            user_id = None
            if user_identifier is not None:
                persisted_user = await data_layer.get_user(user_identifier)
                if persisted_user is None:
                    raise ValueError(f"User '{user_identifier}' not found.")
                user_id = persisted_user.id

            pagination = Pagination(first=first, cursor=cursor)
            filters = ThreadFilter(search=search, userId=user_id)
            threads = await data_layer.list_threads(pagination, filters)
            return self._normalize_threads_tags(threads)

        return self._runtime.run_coroutine_sync(_list_threads())

    def get_thread(self, thread_id: str) -> dict:
        data_layer = self._get_data_layer_or_raise()

        async def _get_thread():
            thread = await data_layer.get_thread(thread_id)
            if thread is None:
                raise ValueError(f"Thread '{thread_id}' not found.")
            return self._normalize_thread_tags(thread)

        return self._runtime.run_coroutine_sync(_get_thread())

    def get_messages(self, thread_id: str) -> dict:
        """Return thread metadata and one ordered list of messages/tool steps."""
        thread = self.get_thread(thread_id)
        return self._build_messages_payload(thread, data_layer=self._data_layer_getter())

    def timeline(self, thread_id: str) -> dict:
        """Backward-compatible alias for `get_messages`."""
        return self.get_messages(thread_id)

    def get_thread_timeline(self, thread_id: str) -> dict:
        """Backward-compatible alias for `get_messages`."""
        return self.get_messages(thread_id)

    def get_thread_messages_and_steps(self, thread_id: str) -> dict:
        """Backward-compatible alias for `get_messages`."""
        return self.get_messages(thread_id)

    def get_timeline(self, thread_id: str) -> dict:
        """Backward-compatible alias for `get_messages`."""
        return self.get_messages(thread_id)

    def update_thread(
        self,
        thread_id: str,
        name: str | None = None,
        metadata: dict | None = None,
        tags: list[str] | None = None,
    ) -> None:
        self._write_thread(
            thread_id=thread_id,
            name=name,
            metadata=metadata,
            tags=tags,
            require_existing=True,
        )

    def new_thread(
        self,
        name: str | None = None,
        metadata: dict | None = None,
        tags: list[str] | None = None,
    ) -> str:
        data_layer = self._get_data_layer_or_raise()
        prepared_tags = self._prepare_tags_for_update(tags, data_layer)

        async def _new_thread() -> str:
            owner_user_id = await self._resolve_default_owner_user_id(data_layer)
            for _ in range(16):
                thread_id = str(self._uuid_factory())
                existing = await data_layer.get_thread(thread_id)
                if existing is not None:
                    continue

                await data_layer.update_thread(
                    thread_id=thread_id,
                    name=name,
                    user_id=owner_user_id,
                    metadata=metadata,
                    tags=prepared_tags,
                )
                return thread_id

            raise RuntimeError("Failed to allocate unique thread_id.")

        return self._runtime.run_coroutine_sync(_new_thread())

    def _write_thread(
        self,
        thread_id: str,
        name: str | None,
        metadata: dict | None,
        tags: list[str] | None,
        require_existing: bool,
    ) -> None:
        data_layer = self._get_data_layer_or_raise()
        prepared_tags = self._prepare_tags_for_update(tags, data_layer)

        async def _write_thread():
            existing = await data_layer.get_thread(thread_id)
            if require_existing and existing is None:
                raise ValueError(f"Thread '{thread_id}' not found.")
            if not require_existing and existing is not None:
                raise ValueError(f"Thread '{thread_id}' already exists.")

            owner_user_id = await self._resolve_default_owner_user_id(data_layer)
            await data_layer.update_thread(
                thread_id=thread_id,
                name=name,
                user_id=owner_user_id,
                metadata=metadata,
                tags=prepared_tags,
            )

        self._runtime.run_coroutine_sync(_write_thread())

    def delete_thread(self, thread_id: str) -> None:
        data_layer = self._get_data_layer_or_raise()

        async def _delete_thread():
            await data_layer.delete_thread(thread_id)

        self._runtime.run_coroutine_sync(_delete_thread())

    def reset_thread(self, thread_id: str) -> None:
        thread = self.get_thread(thread_id)
        thread_name = thread.get("name") if isinstance(thread.get("name"), str) else None
        raw_steps = thread.get("steps")
        step_items = raw_steps if isinstance(raw_steps, list) else []
        step_ids: list[str] = []
        for step in step_items:
            if not isinstance(step, dict):
                continue
            step_id = self._coerce_identifier(step.get("id"))
            if step_id is None:
                continue
            step_ids.append(step_id)

        self._delete_thread_steps_immediately(thread_id=thread_id, step_ids=step_ids)
        self.delete_thread(thread_id)
        self._write_thread(
            thread_id=thread_id,
            name=thread_name,
            metadata=None,
            tags=None,
            require_existing=False,
        )
        self._clear_pending_incoming_for_thread(thread_id)

    def close(self) -> None:
        if self._closed.is_set():
            return

        self._closed.set()
        self._incoming_queue.put_nowait(None)
        self._outgoing_queue.put_nowait(OutgoingCommand(command="close"))

    def is_closed(self) -> bool:
        return self._closed.is_set()

    def _enqueue_incoming(self, message: IncomingMessage) -> None:
        if self._closed.is_set():
            raise AppClosedError("Cannot enqueue incoming message to a closed app.")
        self._incoming_queue.put_nowait(message)

    def _pop_outgoing(self, timeout: float | None = 0.1) -> OutgoingCommand:
        if timeout is None:
            return self._outgoing_queue.get()
        return self._outgoing_queue.get(timeout=timeout)

    def _put_outgoing(self, command: OutgoingCommand) -> None:
        if self._closed.is_set():
            raise AppClosedError("Cannot send command to a closed app.")
        self._outgoing_queue.put_nowait(command)

    def _new_message_id(self) -> str:
        return str(self._uuid_factory())

    def _enqueue_outgoing_command(
        self,
        *,
        command: str,
        thread_id: str,
        message_id: str | None = None,
        content: str | None = None,
        author: str = "Assistant",
        step_type: str | None = None,
        metadata: dict | None = None,
        elements: list[Any] | None = None,
    ) -> None:
        self._put_outgoing(
            OutgoingCommand(
                command=command,
                thread_id=thread_id,
                message_id=message_id,
                content=content,
                elements=elements or [],
                author=author,
                step_type=step_type,
                metadata=metadata or {},
            )
        )

    def _delete_thread_steps_immediately(self, *, thread_id: str, step_ids: list[str]) -> None:
        if not step_ids:
            return

        for step_id in step_ids:
            self._runtime.run_coroutine_sync(
                self._runtime.apply_outgoing_command(
                    OutgoingCommand(
                        command="delete",
                        thread_id=thread_id,
                        message_id=step_id,
                    )
                )
            )

    def _clear_pending_incoming_for_thread(self, thread_id: str) -> None:
        retained_items: list[IncomingMessage | None] = []
        while True:
            try:
                item = self._incoming_queue.get_nowait()
            except queue.Empty:
                break

            if item is None:
                retained_items.append(item)
                continue

            if item.thread_id == thread_id:
                continue

            retained_items.append(item)

        for item in retained_items:
            self._incoming_queue.put_nowait(item)

    def _get_data_layer_or_raise(self):
        data_layer = self._data_layer_getter()
        if data_layer is None:
            raise DataPersistenceNotEnabledError(
                "Data persistence is not enabled. Configure Chainlit data layer first."
            )
        return data_layer

    def _is_sqlite_sqlalchemy_data_layer(self, data_layer: Any) -> bool:
        conninfo = getattr(data_layer, "_conninfo", None)
        if isinstance(conninfo, str) and conninfo.lower().startswith("sqlite"):
            return True

        engine = getattr(data_layer, "engine", None)
        url = getattr(engine, "url", None)
        drivername = getattr(url, "drivername", None)
        return isinstance(drivername, str) and drivername.lower().startswith("sqlite")

    def _prepare_tags_for_update(
        self,
        tags: list[str] | None,
        data_layer: Any,
    ) -> list[str] | str | None:
        if tags is None:
            return None
        if self._is_sqlite_sqlalchemy_data_layer(data_layer):
            return json.dumps(tags)
        return tags

    async def _resolve_default_owner_user_id(self, data_layer: Any) -> str | None:
        auth = self._runtime.get_auth()
        if auth is None:
            return None

        identifier = auth.identifier or auth.username
        persisted_user = await data_layer.get_user(identifier)
        if persisted_user is not None:
            return persisted_user.id

        if not hasattr(data_layer, "create_user"):
            return None

        created_user = await data_layer.create_user(
            User(
                identifier=identifier,
                metadata=auth.metadata or {},
            )
        )
        if created_user is None:
            return None

        return created_user.id

    def _normalize_thread_tags(self, thread: dict) -> dict:
        tags = thread.get("tags")
        if not isinstance(tags, str):
            return thread

        try:
            decoded = json.loads(tags)
        except json.JSONDecodeError:
            return thread

        if not isinstance(decoded, list):
            return thread

        normalized = dict(thread)
        normalized["tags"] = decoded
        return normalized

    def _normalize_threads_tags(self, threads):
        if not hasattr(threads, "data"):
            return threads

        threads.data = [
            self._normalize_thread_tags(thread)
            if isinstance(thread, dict)
            else thread
            for thread in threads.data
        ]
        return threads

    def _build_messages_payload(self, thread: dict, *, data_layer: Any | None = None) -> dict:
        raw_steps = thread.get("steps")
        step_items = raw_steps if isinstance(raw_steps, list) else []
        messages = self._filter_message_steps(step_items)

        raw_elements = thread.get("elements")
        element_items = raw_elements if isinstance(raw_elements, list) else []
        elements_by_for_id = self._index_elements_by_for_id(element_items)

        enriched_messages: list[dict] = []
        for message in messages:
            message_copy = dict(message)
            message_id = self._coerce_identifier(message_copy.get("id"))
            if message_id is not None:
                related_elements = elements_by_for_id.get(message_id, [])
                message_copy["elements"] = [
                    self._normalize_message_element(element, data_layer=data_layer)
                    for element in related_elements
                ]
            else:
                message_copy["elements"] = []
            enriched_messages.append(message_copy)

        thread_metadata = dict(thread)
        thread_metadata.pop("steps", None)

        return {
            "thread": thread_metadata,
            "messages": enriched_messages,
        }

    def _filter_message_steps(self, steps: list[dict]) -> list[dict]:
        filtered_steps: list[dict] = []
        for step in steps:
            if not isinstance(step, dict):
                continue
            step_type = step.get("type")
            if not isinstance(step_type, str):
                continue
            if step_type in self._MESSAGE_STEP_TYPES:
                filtered_steps.append(step)
        return filtered_steps

    def _index_elements_by_for_id(self, elements: list[dict]) -> dict[str, list[dict]]:
        elements_by_for_id: dict[str, list[dict]] = {}
        for element in elements:
            if not isinstance(element, dict):
                continue
            for_id = self._extract_element_target_id(element)
            if for_id is None:
                continue
            elements_by_for_id.setdefault(for_id, []).append(element)
        return elements_by_for_id

    def _extract_element_target_id(self, element: dict) -> str | None:
        for key in ("forId", "for_id", "stepId", "step_id"):
            identifier = self._coerce_identifier(element.get(key))
            if identifier is not None:
                return identifier
        return None

    def _normalize_message_element(self, element: dict, *, data_layer: Any | None = None) -> dict:
        normalized = dict(element)
        source = self._resolve_element_source(normalized, data_layer=data_layer)
        normalized["source"] = source
        normalized["has_source"] = source is not None
        return normalized

    def _resolve_element_source(
        self,
        element: dict,
        *,
        data_layer: Any | None = None,
    ) -> dict[str, Any] | None:
        object_key = self._coerce_non_empty_string(
            element.get("objectKey") or element.get("object_key")
        )
        if object_key is not None:
            resolved_url = self._resolve_element_url_from_object_key(
                object_key=object_key,
                data_layer=data_layer,
            )
            resolved_path = self._resolve_element_path_from_object_key(
                object_key=object_key,
                data_layer=data_layer,
            )
            if resolved_path is not None:
                element["path"] = resolved_path
                if resolved_url is not None:
                    element["url"] = resolved_url
                return {"kind": "path", "value": resolved_path}

            if resolved_url is not None:
                element["url"] = resolved_url
                return {"kind": "url", "value": resolved_url}

        path = self._coerce_non_empty_string(element.get("path"))
        if path is not None:
            return {"kind": "path", "value": path}

        url = self._coerce_non_empty_string(element.get("url"))
        if url is not None:
            return {"kind": "url", "value": url}

        content = element.get("content")
        if isinstance(content, (bytes, bytearray)):
            return {"kind": "bytes", "value": {"length": len(content)}}
        if isinstance(content, str) and content:
            return {"kind": "bytes", "value": {"length": len(content.encode("utf-8"))}}

        if object_key is not None:
            return {"kind": "objectKey", "value": object_key}

        chainlit_key = self._coerce_non_empty_string(
            element.get("chainlitKey") or element.get("chainlit_key")
        )
        if chainlit_key is not None:
            return {"kind": "chainlitKey", "value": chainlit_key}

        return None

    def _resolve_element_url_from_object_key(
        self,
        *,
        object_key: str,
        data_layer: Any | None = None,
    ) -> str | None:
        storage = self._resolve_storage_provider(data_layer)
        if storage is None:
            return None

        get_read_url = getattr(storage, "get_read_url", None)
        if not callable(get_read_url):
            return None

        try:
            maybe_url = get_read_url(object_key)
            if inspect.isawaitable(maybe_url):
                maybe_url = self._runtime.run_coroutine_sync(maybe_url)
        except Exception:
            return None

        if isinstance(maybe_url, str) and maybe_url.strip():
            return maybe_url
        return None

    def _resolve_element_path_from_object_key(
        self,
        *,
        object_key: str,
        data_layer: Any | None = None,
    ) -> str | None:
        storage = self._resolve_storage_provider(data_layer)
        if storage is None:
            return None

        resolve_file_path = getattr(storage, "resolve_file_path", None)
        if not callable(resolve_file_path):
            return None

        try:
            maybe_path = resolve_file_path(object_key)
        except Exception:
            return None

        if maybe_path is None:
            return None

        path = maybe_path if isinstance(maybe_path, Path) else Path(str(maybe_path))
        if not path.is_file():
            return None
        return str(path)

    def _resolve_storage_provider(self, data_layer: Any | None):
        if data_layer is None:
            return None

        for attr in ("storage_provider", "storage_client"):
            storage = getattr(data_layer, attr, None)
            if storage is not None:
                return storage
        return None

    def _coerce_non_empty_string(self, value: Any) -> str | None:
        if isinstance(value, str) and value:
            return value
        return None

    def _coerce_identifier(self, value: Any) -> str | None:
        if isinstance(value, str):
            return value if value else None
        if value is None:
            return None
        if isinstance(value, (dict, list, tuple, set, bytes, bytearray)):
            return None

        rendered = str(value)
        if not rendered:
            return None
        return rendered



from __future__ import annotations

import asyncio
import json
import logging
import mimetypes
import queue
import re
import threading
from collections.abc import Awaitable, Callable
from pathlib import Path
from typing import TYPE_CHECKING, Any, Coroutine
from urllib.parse import unquote, urlparse
from uuid import uuid4

import aiohttp
from chainlit.context import init_http_context
from chainlit.data import get_data_layer
from chainlit.utils import utc_now

from .discord_outgoing import send_discord_command, supports_discord_command
from .errors import ThreadSessionNotActiveError
from .models import IncomingMessage, OutgoingCommand
from .storage.local import LOCAL_STORAGE_ROUTE_PREFIX

if TYPE_CHECKING:
    from chainlit.session import WebsocketSession

    from .app import EasierlitApp
    from .client import EasierlitClient
    from .settings import EasierlitAuthConfig, EasierlitPersistenceConfig
    from .storage.local import LocalFileStorageClient

LOGGER = logging.getLogger(__name__)

_ELEMENT_DB_COLUMNS = (
    "id",
    "threadId",
    "type",
    "chainlitKey",
    "url",
    "objectKey",
    "name",
    "display",
    "size",
    "language",
    "page",
    "autoPlay",
    "playerConfig",
    "forId",
    "mime",
    "props",
)


class RuntimeRegistry:
    def __init__(
        self,
        *,
        data_layer_getter: Callable[[], Any | None] = get_data_layer,
        init_http_context_fn: Callable[..., Any] = init_http_context,
        utc_now_fn: Callable[[], Any] = utc_now,
    ) -> None:
        self._client: EasierlitClient | None = None
        self._app: EasierlitApp | None = None
        self._auth: EasierlitAuthConfig | None = None
        self._persistence: EasierlitPersistenceConfig | None = None
        self._discord_token: str | None = None

        self._thread_to_session: dict[str, str] = {}
        self._session_to_thread: dict[str, str] = {}
        self._thread_to_discord_channel: dict[str, int] = {}

        self._main_loop: asyncio.AbstractEventLoop | None = None
        self._dispatcher_task: asyncio.Task[None] | None = None

        self._data_layer_getter = data_layer_getter
        self._init_http_context_fn = init_http_context_fn
        self._utc_now_fn = utc_now_fn
        self._discord_sender: Callable[[int, OutgoingCommand], Awaitable[bool]] | None = None

        self._lock = threading.RLock()

    def bind(
        self,
        *,
        client: EasierlitClient,
        app: EasierlitApp,
        auth: EasierlitAuthConfig | None = None,
        persistence: EasierlitPersistenceConfig | None = None,
        discord_token: str | None = None,
    ) -> None:
        with self._lock:
            self._client = client
            self._app = app
            self._auth = auth
            self._persistence = persistence
            self._discord_token = discord_token
            self._thread_to_session.clear()
            self._session_to_thread.clear()
            self._thread_to_discord_channel.clear()
            self._discord_sender = None

    def unbind(self) -> None:
        with self._lock:
            self._client = None
            self._app = None
            self._auth = None
            self._persistence = None
            self._discord_token = None
            self._thread_to_session.clear()
            self._session_to_thread.clear()
            self._thread_to_discord_channel.clear()
            self._main_loop = None
            self._dispatcher_task = None
            self._discord_sender = None

    def get_client(self) -> EasierlitClient | None:
        return self._client

    def get_app(self) -> EasierlitApp | None:
        return self._app

    def get_auth(self) -> EasierlitAuthConfig | None:
        return self._auth

    def get_persistence(self) -> EasierlitPersistenceConfig | None:
        return self._persistence

    def get_discord_token(self) -> str | None:
        return self._discord_token

    def set_discord_sender(
        self,
        sender: Callable[[int, OutgoingCommand], Awaitable[bool]] | None,
    ) -> None:
        with self._lock:
            self._discord_sender = sender

    def set_main_loop(self, loop: asyncio.AbstractEventLoop) -> None:
        self._main_loop = loop

    def get_main_loop(self) -> asyncio.AbstractEventLoop | None:
        return self._main_loop

    def run_coroutine_sync(self, coro: Coroutine[Any, Any, Any]) -> Any:
        loop = self._main_loop
        if loop and loop.is_running():
            try:
                running_loop = asyncio.get_running_loop()
            except RuntimeError:
                running_loop = None

            if running_loop is loop:
                raise RuntimeError(
                    "Cannot synchronously wait for a coroutine from the Chainlit event loop."
                )

            future = asyncio.run_coroutine_threadsafe(coro, loop)
            return future.result()

        return asyncio.run(coro)

    def register_session(self, thread_id: str, session_id: str) -> None:
        with self._lock:
            self._thread_to_session[thread_id] = session_id
            self._session_to_thread[session_id] = thread_id

    def register_discord_channel(self, thread_id: str, channel_id: int) -> None:
        with self._lock:
            self._thread_to_discord_channel[thread_id] = channel_id

    def unregister_session(self, session_id: str) -> None:
        with self._lock:
            thread_id = self._session_to_thread.pop(session_id, None)
            if thread_id is not None:
                self._thread_to_session.pop(thread_id, None)

    def get_session_id_for_thread(self, thread_id: str) -> str | None:
        with self._lock:
            return self._thread_to_session.get(thread_id)

    def get_discord_channel_for_thread(self, thread_id: str) -> int | None:
        with self._lock:
            return self._thread_to_discord_channel.get(thread_id)

    def enqueue_incoming(self, message: IncomingMessage) -> None:
        app = self._app
        if app is None:
            return
        app._enqueue_incoming(message)

    async def start_dispatcher(self) -> None:
        if self._app is None:
            raise RuntimeError("No active app bound to runtime.")

        if self._dispatcher_task and not self._dispatcher_task.done():
            return

        self._dispatcher_task = asyncio.create_task(self._dispatch_outgoing_loop())

    async def stop_dispatcher(self) -> None:
        task = self._dispatcher_task
        if task is None:
            return

        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass
        finally:
            self._dispatcher_task = None

    async def _dispatch_outgoing_loop(self) -> None:
        app = self._app
        if app is None:
            return

        while not app.is_closed():
            try:
                command = await asyncio.to_thread(app._pop_outgoing, 0.2)
            except queue.Empty:
                continue

            if command.command == "close":
                break

            try:
                await self.apply_outgoing_command(command)
            except Exception:
                LOGGER.exception(
                    "Failed to apply outgoing command: %s",
                    command.model_dump(),
                )

    def _resolve_session(self, thread_id: str) -> WebsocketSession | None:
        from chainlit.session import WebsocketSession

        session_id = self.get_session_id_for_thread(thread_id)
        if session_id is None:
            return None

        session = WebsocketSession.get_by_id(session_id)
        if session is None:
            self.unregister_session(session_id)
            return None

        return session

    @staticmethod
    def _require_thread_id(command: OutgoingCommand) -> str:
        thread_id = command.thread_id
        if not thread_id:
            raise ValueError("Outgoing command is missing thread_id.")
        return thread_id

    @staticmethod
    def _require_message_id(command: OutgoingCommand, *, action: str) -> str:
        message_id = command.message_id
        if not message_id:
            raise ValueError(f"{action} command requires message_id.")
        return message_id

    @staticmethod
    def _is_tool_command(command_name: str) -> bool:
        return command_name in ("add_tool", "update_tool")

    @staticmethod
    def _is_create_command(command_name: str) -> bool:
        return command_name in ("add_message", "add_tool")

    @staticmethod
    def _is_update_command(command_name: str) -> bool:
        return command_name in ("update_message", "update_tool")

    async def apply_outgoing_command(self, command: OutgoingCommand) -> None:
        if command.command == "close":
            return

        thread_id = self._require_thread_id(command)
        command = await self._prepare_command_elements(command, thread_id=thread_id)

        session_handled = False
        session = self._resolve_session(thread_id)
        if session is not None:
            await self._apply_realtime_command(session, command)
            session_handled = True

        discord_handled = False
        discord_channel_id = self.get_discord_channel_for_thread(thread_id)
        if discord_channel_id is not None:
            discord_handled = await self._apply_discord_command(discord_channel_id, command)

        data_layer = self._data_layer_getter()
        if data_layer:
            await self._apply_data_layer_command(command)
            return

        if session_handled or discord_handled:
            return

        if not session_handled and not discord_handled:
            raise ThreadSessionNotActiveError(
                f"Thread '{thread_id}' has no active session and no data layer fallback."
            )

    async def _prepare_command_elements(
        self,
        command: OutgoingCommand,
        *,
        thread_id: str,
    ) -> OutgoingCommand:
        if command.command == "delete":
            return command
        if not command.elements:
            return command
        if self.__resolve_local_storage_provider() is None:
            return command

        message_id = self._require_message_id(command, action=command.command)
        prepared: list[dict[str, Any]] = []
        for index, element in enumerate(command.elements):
            normalized = await self._prepare_element_for_local_storage(
                element=element,
                thread_id=thread_id,
                message_id=message_id,
                index=index,
            )
            if normalized is None:
                continue
            prepared.append(normalized)

        return command.model_copy(update={"elements": prepared})

    async def _apply_realtime_command(
        self, session: WebsocketSession, command: OutgoingCommand
    ) -> None:
        from chainlit.context import context, init_ws_context

        init_ws_context(session)
        thread_id = self._require_thread_id(command)

        if command.command == "add_message":
            message_id = self._require_message_id(command, action="Add message")
            step_dict = self._build_step_payload(
                command=command,
                thread_id=thread_id,
                message_id=message_id,
            )
            await context.emitter.send_step(step_dict)
            await self._emit_realtime_elements(command.elements)
            return

        if command.command == "update_message":
            message_id = self._require_message_id(command, action="Update")
            step_dict = self._build_step_payload(
                command=command,
                thread_id=thread_id,
                message_id=message_id,
            )
            await context.emitter.update_step(step_dict)
            await self._emit_realtime_elements(command.elements)
            return

        if command.command == "add_tool":
            message_id = self._require_message_id(command, action="Add tool")
            step_dict = self._build_step_payload(
                command=command,
                thread_id=thread_id,
                message_id=message_id,
            )
            await context.emitter.send_step(step_dict)
            await self._emit_realtime_elements(command.elements)
            return

        if command.command == "update_tool":
            message_id = self._require_message_id(command, action="Update tool")
            step_dict = self._build_step_payload(
                command=command,
                thread_id=thread_id,
                message_id=message_id,
            )
            await context.emitter.update_step(step_dict)
            await self._emit_realtime_elements(command.elements)
            return

        if command.command == "delete":
            message_id = self._require_message_id(command, action="Delete")
            await context.emitter.delete_step({"id": message_id, "threadId": thread_id})
            return

        raise ValueError(f"Unsupported command: {command.command}")

    async def _emit_realtime_elements(self, elements: list[Any]) -> None:
        if not elements:
            return

        from chainlit.context import context

        for element in elements:
            payload = element if isinstance(element, dict) else self._coerce_element_dict(element)
            if not isinstance(payload, dict):
                continue
            await context.emitter.send_element(payload)

    def _build_step_payload(
        self,
        *,
        command: OutgoingCommand,
        thread_id: str,
        message_id: str,
        timestamp: Any | None = None,
    ) -> dict[str, Any]:
        created_at = timestamp if timestamp is not None else self._utc_now_fn()
        is_tool_command = self._is_tool_command(command.command)
        step_type = command.step_type or ("tool" if is_tool_command else "assistant_message")
        return {
            "id": message_id,
            "threadId": thread_id,
            "name": command.author,
            "type": step_type,
            "output": command.content or "",
            "createdAt": created_at,
            "start": created_at,
            "end": created_at,
            "streaming": False,
            "isError": False,
            "waitForAnswer": False,
            "metadata": command.metadata or {},
        }

    async def _apply_data_layer_command(self, command: OutgoingCommand) -> None:
        data_layer = self._data_layer_getter()
        if not data_layer:
            raise RuntimeError("Data layer unexpectedly missing.")

        thread_id = self._require_thread_id(command)
        self._init_data_layer_http_context(thread_id)

        if command.command == "delete":
            message_id = self._require_message_id(command, action="Delete")
            await data_layer.delete_step(message_id)
            return

        message_id = self._require_message_id(command, action=command.command)
        timestamp = self._utc_now_fn()
        step_dict = self._build_step_payload(
            command=command,
            thread_id=thread_id,
            message_id=message_id,
            timestamp=timestamp,
        )

        if self._is_create_command(command.command):
            await data_layer.create_step(step_dict)
            await self._persist_data_layer_elements(
                data_layer=data_layer,
                thread_id=thread_id,
                message_id=message_id,
                elements=command.elements,
            )
            return

        if self._is_update_command(command.command):
            await data_layer.update_step(step_dict)
            await self._persist_data_layer_elements(
                data_layer=data_layer,
                thread_id=thread_id,
                message_id=message_id,
                elements=command.elements,
            )
            return

        raise ValueError(f"Unsupported command: {command.command}")

    async def _persist_data_layer_elements(
        self,
        *,
        data_layer: Any,
        thread_id: str,
        message_id: str,
        elements: list[Any],
    ) -> None:
        if not elements:
            return

        if self._can_upsert_element_directly(data_layer):
            for element in elements:
                element_dict = self._prepare_element_record(
                    element=element,
                    thread_id=thread_id,
                    message_id=message_id,
                )
                if element_dict is None:
                    continue
                await self._upsert_sqlalchemy_element(data_layer=data_layer, element_dict=element_dict)
            return

        create_element = getattr(data_layer, "create_element", None)
        if not callable(create_element):
            return

        for element in elements:
            if hasattr(element, "for_id"):
                element.for_id = message_id
            if hasattr(element, "thread_id"):
                element.thread_id = thread_id
            await create_element(element)

    def _can_upsert_element_directly(self, data_layer: Any) -> bool:
        execute_sql = getattr(data_layer, "execute_sql", None)
        return callable(execute_sql)

    async def _upsert_sqlalchemy_element(
        self,
        *,
        data_layer: Any,
        element_dict: dict[str, Any],
    ) -> None:
        execute_sql = getattr(data_layer, "execute_sql", None)
        if not callable(execute_sql):
            return

        payload = {
            column: element_dict.get(column)
            for column in _ELEMENT_DB_COLUMNS
            if element_dict.get(column) is not None
        }
        if not payload.get("id"):
            return

        props = payload.get("props")
        if isinstance(props, dict):
            payload["props"] = json.dumps(props, ensure_ascii=False)
        elif props is not None and not isinstance(props, str):
            payload["props"] = json.dumps(props, ensure_ascii=False)

        columns = ", ".join(f'"{column}"' for column in payload.keys())
        placeholders = ", ".join(f":{column}" for column in payload.keys())
        updates = ", ".join(
            f'"{column}" = :{column}' for column in payload.keys() if column != "id"
        )
        if updates:
            query = (
                f"INSERT INTO elements ({columns}) VALUES ({placeholders}) "
                f"ON CONFLICT (id) DO UPDATE SET {updates};"
            )
        else:
            query = (
                f"INSERT INTO elements ({columns}) VALUES ({placeholders}) "
                'ON CONFLICT (id) DO NOTHING;'
            )
        await execute_sql(query=query, parameters=payload)

    async def _prepare_element_for_local_storage(
        self,
        *,
        element: Any,
        thread_id: str,
        message_id: str,
        index: int,
    ) -> dict[str, Any] | None:
        local_storage = self.__resolve_local_storage_provider()
        if local_storage is None:
            return self._prepare_element_record(
                element=element,
                thread_id=thread_id,
                message_id=message_id,
            )

        element_dict = self._coerce_element_dict(element)
        element_id = self._coerce_text(element_dict.get("id")) or str(uuid4())
        element_name = self._coerce_text(element_dict.get("name")) or f"element-{index + 1}"
        mime = self._coerce_text(element_dict.get("mime")) or self._guess_mime_type(element_name)
        element_type = self._coerce_text(element_dict.get("type")) or self._infer_element_type_from_mime(
            mime
        )
        object_key = self._resolve_element_object_key(element_dict)
        url = self._coerce_text(element_dict.get("url"))

        async def _resolve_uploaded_reference(
            uploaded: dict[str, Any],
            *,
            fallback_object_key: str,
        ) -> tuple[str, str | None]:
            resolved_object_key = (
                self._coerce_text(uploaded.get("object_key")) or fallback_object_key
            )
            uploaded_url = self._coerce_text(uploaded.get("url"))
            if uploaded_url:
                return resolved_object_key, uploaded_url
            try:
                return resolved_object_key, await local_storage.get_read_url(resolved_object_key)
            except Exception:
                return resolved_object_key, None

        if object_key:
            try:
                url = await local_storage.get_read_url(object_key)
            except Exception:
                payload = await self._resolve_element_payload(element_dict, local_storage=local_storage)
                if payload is None:
                    url = None
                else:
                    try:
                        uploaded = await local_storage.upload_file(
                            object_key=object_key,
                            data=payload,
                            mime=mime,
                            overwrite=True,
                        )
                        object_key, url = await _resolve_uploaded_reference(
                            uploaded, fallback_object_key=object_key
                        )
                    except Exception:
                        url = None
        else:
            payload = await self._resolve_element_payload(element_dict, local_storage=local_storage)
            if payload is None:
                LOGGER.warning(
                    "Skipping element '%s' for message '%s': no readable source.",
                    element_name,
                    message_id,
                )
                return None
            object_key = self._build_generated_object_key(
                thread_id=thread_id,
                message_id=message_id,
                element_id=element_id,
                element_name=element_name,
            )
            uploaded = await local_storage.upload_file(
                object_key=object_key,
                data=payload,
                mime=mime,
                overwrite=True,
            )
            object_key, url = await _resolve_uploaded_reference(
                uploaded, fallback_object_key=object_key
            )

        if not object_key and not url:
            LOGGER.warning(
                "Skipping element '%s' for message '%s': failed to resolve objectKey/url.",
                element_name,
                message_id,
            )
            return None

        normalized = {
            "id": element_id,
            "threadId": thread_id,
            "type": element_type,
            "url": url,
            "objectKey": object_key,
            "name": element_name,
            "display": self._coerce_text(element_dict.get("display")) or "inline",
            "size": self._coerce_text(element_dict.get("size")),
            "language": self._coerce_text(element_dict.get("language")),
            "page": element_dict.get("page"),
            "autoPlay": element_dict.get("autoPlay"),
            "playerConfig": element_dict.get("playerConfig"),
            "forId": message_id,
            "mime": mime,
            "props": element_dict.get("props"),
        }

        return {key: value for key, value in normalized.items() if value is not None}

    async def _resolve_element_payload(
        self,
        element_dict: dict[str, Any],
        *,
        local_storage: "LocalFileStorageClient",
    ) -> bytes | None:
        path = self._coerce_text(element_dict.get("path"))
        if path:
            file_path = Path(path).expanduser()
            if not file_path.is_absolute():
                file_path = Path.cwd() / file_path
            if file_path.is_file():
                return file_path.read_bytes()
            return None

        content = element_dict.get("content")
        if isinstance(content, (bytes, bytearray)):
            return bytes(content)
        if isinstance(content, str):
            return content.encode("utf-8")

        url = self._coerce_text(element_dict.get("url"))
        if not url:
            return None

        local_object_key = self._extract_local_route_object_key(url)
        if local_object_key:
            try:
                local_file_path = local_storage.resolve_file_path(local_object_key)
            except Exception:
                local_file_path = None
            if local_file_path is not None and local_file_path.is_file():
                return local_file_path.read_bytes()

        return await self._download_url_bytes(url)

    async def _download_url_bytes(self, url: str) -> bytes | None:
        parsed = urlparse(url)
        if parsed.scheme not in ("http", "https"):
            return None
        try:
            async with aiohttp.ClientSession() as client:
                async with client.get(url) as response:
                    if response.status != 200:
                        return None
                    return await response.read()
        except Exception:
            LOGGER.exception("Failed to download element URL: %s", url)
            return None

    def __resolve_local_storage_provider(self) -> "LocalFileStorageClient" | None:
        persistence = self._persistence
        if persistence is None or not persistence.enabled:
            return None

        from .settings import _resolve_local_storage_provider

        try:
            return _resolve_local_storage_provider(persistence)
        except (TypeError, ValueError):
            return None

    def _prepare_element_record(
        self,
        *,
        element: Any,
        thread_id: str,
        message_id: str,
    ) -> dict[str, Any] | None:
        if not isinstance(element, dict):
            element = self._coerce_element_dict(element)
        if not isinstance(element, dict):
            return None

        element_id = self._coerce_text(element.get("id")) or str(uuid4())

        normalized = dict(element)
        normalized["id"] = element_id
        normalized["threadId"] = self._coerce_text(normalized.get("threadId")) or thread_id
        normalized["forId"] = self._coerce_text(normalized.get("forId")) or message_id
        normalized["type"] = self._coerce_text(normalized.get("type")) or "file"
        normalized["display"] = self._coerce_text(normalized.get("display")) or "inline"
        normalized["name"] = self._coerce_text(normalized.get("name")) or "file"

        if "object_key" in normalized and "objectKey" not in normalized:
            normalized["objectKey"] = normalized.get("object_key")
        if "chainlit_key" in normalized and "chainlitKey" not in normalized:
            normalized["chainlitKey"] = normalized.get("chainlit_key")
        if "for_id" in normalized and "forId" not in normalized:
            normalized["forId"] = normalized.get("for_id")
        if "thread_id" in normalized and "threadId" not in normalized:
            normalized["threadId"] = normalized.get("thread_id")

        normalized.pop("path", None)
        normalized.pop("content", None)

        return {key: value for key, value in normalized.items() if value is not None}

    def _coerce_element_dict(self, element: Any) -> dict[str, Any]:
        if isinstance(element, dict):
            return dict(element)

        result: dict[str, Any] = {}
        to_dict = getattr(element, "to_dict", None)
        if callable(to_dict):
            try:
                dumped = to_dict()
            except Exception:
                dumped = None
            if isinstance(dumped, dict):
                result.update(dumped)

        attr_map = {
            "id": "id",
            "threadId": "thread_id",
            "type": "type",
            "chainlitKey": "chainlit_key",
            "url": "url",
            "objectKey": "object_key",
            "path": "path",
            "content": "content",
            "name": "name",
            "display": "display",
            "size": "size",
            "language": "language",
            "page": "page",
            "autoPlay": "auto_play",
            "playerConfig": "player_config",
            "forId": "for_id",
            "mime": "mime",
            "props": "props",
        }
        for key, attr_name in attr_map.items():
            if key in result:
                continue
            if hasattr(element, key):
                result[key] = getattr(element, key)
                continue
            if hasattr(element, attr_name):
                result[key] = getattr(element, attr_name)

        return result

    def _resolve_element_object_key(self, element_dict: dict[str, Any]) -> str | None:
        object_key = self._coerce_text(element_dict.get("objectKey"))
        if object_key:
            return object_key
        object_key = self._coerce_text(element_dict.get("object_key"))
        if object_key:
            return object_key
        url = self._coerce_text(element_dict.get("url"))
        if not url:
            return None
        return self._extract_local_route_object_key(url)

    def _extract_local_route_object_key(self, url: str) -> str | None:
        path = urlparse(url).path or ""
        prefix = f"{LOCAL_STORAGE_ROUTE_PREFIX}/"
        marker = path.find(prefix)
        if marker < 0:
            return None
        object_key = path[marker + len(prefix) :]
        if not object_key:
            return None
        decoded = unquote(object_key).strip("/")
        return decoded or None

    def _build_generated_object_key(
        self,
        *,
        thread_id: str,
        message_id: str,
        element_id: str,
        element_name: str,
    ) -> str:
        safe_thread = self._safe_path_segment(thread_id)
        safe_message = self._safe_path_segment(message_id)
        safe_element = self._safe_path_segment(element_id)
        safe_name = self._safe_file_name(element_name)
        return f"{safe_thread}/{safe_message}/{safe_element}/{safe_name}"

    def _safe_path_segment(self, value: str) -> str:
        rendered = self._coerce_text(value) or "item"
        sanitized = re.sub(r"[^A-Za-z0-9._-]+", "-", rendered).strip("-._")
        return sanitized or "item"

    def _safe_file_name(self, value: str) -> str:
        raw_name = Path(value).name
        if not raw_name:
            raw_name = "file"
        sanitized = re.sub(r"[^A-Za-z0-9._-]+", "-", raw_name).strip("-._")
        return sanitized or "file"

    def _guess_mime_type(self, name: str) -> str:
        guessed, _ = mimetypes.guess_type(name)
        if guessed:
            return guessed
        return "application/octet-stream"

    def _infer_element_type_from_mime(self, mime: str) -> str:
        normalized = (mime or "").lower()
        if normalized.startswith("image/"):
            return "image"
        if normalized.startswith("audio/"):
            return "audio"
        if normalized.startswith("video/"):
            return "video"
        if normalized == "application/pdf":
            return "pdf"
        if normalized.startswith("text/"):
            return "text"
        return "file"

    def _coerce_text(self, value: Any) -> str | None:
        if isinstance(value, str):
            stripped = value.strip()
            return stripped or None
        if value is None:
            return None
        if isinstance(value, (bytes, bytearray, dict, list, tuple, set)):
            return None
        rendered = str(value).strip()
        return rendered or None

    async def _apply_discord_command(self, channel_id: int, command: OutgoingCommand) -> bool:
        if not supports_discord_command(command.command):
            return False

        discord_sender = self._discord_sender
        if discord_sender is not None:
            try:
                return await discord_sender(channel_id, command)
            except Exception:
                LOGGER.exception("Registered Discord sender failed for channel %s.", channel_id)
                return False

        try:
            from chainlit.discord.app import client
        except Exception:
            LOGGER.exception("Failed to import Chainlit Discord client.")
            return False

        return await send_discord_command(
            client=client,
            channel_id=channel_id,
            command=command,
            logger=LOGGER,
        )

    def _init_data_layer_http_context(self, thread_id: str) -> None:
        self._init_http_context_fn(thread_id=thread_id, client_type="webapp")


_RUNTIME = RuntimeRegistry()


def get_runtime() -> RuntimeRegistry:
    return _RUNTIME



import asyncio
from contextlib import contextmanager
import os
from types import SimpleNamespace

import chainlit.data.sql_alchemy as chainlit_sql_alchemy
from chainlit.config import config
import pytest
from fastapi.testclient import TestClient

import easierlit.chainlit_entry as chainlit_entry
from easierlit import EasierlitApp, EasierlitClient, EasierlitPersistenceConfig
from easierlit.errors import AppClosedError, RunFuncExecutionError
from easierlit.runtime import get_runtime
from easierlit.settings import _resolve_local_storage_provider
from easierlit.storage import LocalFileStorageClient


class _FakeUIMessage:
    sent: list[tuple[str, str | None]] = []

    def __init__(self, content: str, author: str | None = None):
        self.content = content
        self.author = author

    async def send(self):
        self.__class__.sent.append((self.content, self.author))
        return self


@contextmanager
def _swap_attr(obj, name: str, value):
    original = getattr(obj, name)
    setattr(obj, name, value)
    try:
        yield
    finally:
        setattr(obj, name, original)


@pytest.fixture(autouse=True)
def _reset_runtime_state():
    runtime = get_runtime()
    runtime.unbind()

    chainlit_entry._APP_CLOSED_WARNING_EMITTED = False
    chainlit_entry._WORKER_FAILURE_UI_NOTIFIED = False
    chainlit_entry._DISCORD_BRIDGE = None
    chainlit_entry._DEFAULT_DATA_LAYER_REGISTERED = False
    chainlit_entry._CONFIG_APPLIED = False
    chainlit_entry._LOCAL_STORAGE_PROVIDER = None

    yield

    runtime.unbind()
    chainlit_entry._APP_CLOSED_WARNING_EMITTED = False
    chainlit_entry._WORKER_FAILURE_UI_NOTIFIED = False
    chainlit_entry._DISCORD_BRIDGE = None
    chainlit_entry._DEFAULT_DATA_LAYER_REGISTERED = False
    chainlit_entry._CONFIG_APPLIED = False
    chainlit_entry._LOCAL_STORAGE_PROVIDER = None


def _sample_message():
    return SimpleNamespace(
        id="msg-1",
        content="/help",
        elements=[],
        author="User",
        created_at=None,
        metadata=None,
    )


def test_on_message_swallows_closed_app_with_worker_error(caplog):
    runtime = get_runtime()
    app = EasierlitApp()
    app.close()
    client = EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread")
    runtime.bind(client=client, app=app)
    client._record_worker_error("Traceback (most recent call last):\nRuntimeError: boom")

    _FakeUIMessage.sent = []
    fake_session = SimpleNamespace(thread_id="thread-1", id="session-1")

    caplog.set_level("WARNING")
    with _swap_attr(chainlit_entry.cl, "context", SimpleNamespace(session=fake_session)):
        with _swap_attr(chainlit_entry.cl, "Message", _FakeUIMessage):
            asyncio.run(chainlit_entry._on_message(_sample_message()))
            asyncio.run(chainlit_entry._on_message(_sample_message()))

    assert len(_FakeUIMessage.sent) == 1
    assert "Server is shutting down" in _FakeUIMessage.sent[0][0]
    assert "server shutdown in progress" in caplog.text.lower()


def test_on_message_raises_when_closed_without_worker_error():
    runtime = get_runtime()
    app = EasierlitApp()
    app.close()
    client = EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread")
    runtime.bind(client=client, app=app)

    fake_session = SimpleNamespace(thread_id="thread-1", id="session-1")
    with _swap_attr(chainlit_entry.cl, "context", SimpleNamespace(session=fake_session)):
        with pytest.raises(AppClosedError):
            asyncio.run(chainlit_entry._on_message(_sample_message()))


def test_on_app_shutdown_logs_summary_without_traceback(caplog):
    runtime = get_runtime()
    app = EasierlitApp()
    client = EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread")
    runtime.bind(client=client, app=app)

    worker_traceback = (
        "Traceback (most recent call last):\n"
        "RuntimeError: command '/new' failed: Chainlit context not found"
    )
    client._record_worker_error(worker_traceback)

    def fake_stop():
        raise RunFuncExecutionError(worker_traceback)

    original_stop = client.stop
    client.stop = fake_stop

    caplog.set_level("WARNING")
    try:
        asyncio.run(chainlit_entry._on_app_shutdown())
    finally:
        client.stop = original_stop

    assert "run_func crash acknowledged during shutdown" in caplog.text
    assert "command '/new' failed: Chainlit context not found" in caplog.text
    assert "Traceback (most recent call last)" not in caplog.text


def test_on_message_registers_discord_channel():
    runtime = get_runtime()
    app = EasierlitApp()
    client = EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread")
    runtime.bind(client=client, app=app)

    fake_session = SimpleNamespace(thread_id="thread-1", id="session-1", client_type="discord")
    fake_channel = SimpleNamespace(id=777)
    fake_user_session = SimpleNamespace(get=lambda key: fake_channel if key == "discord_channel" else None)

    captured: dict[str, int | str] = {}
    original_register_discord_channel = runtime.register_discord_channel

    def fake_register_discord_channel(thread_id: str, channel_id: int):
        captured["thread_id"] = thread_id
        captured["channel_id"] = channel_id

    runtime.register_discord_channel = fake_register_discord_channel

    try:
        with _swap_attr(chainlit_entry.cl, "context", SimpleNamespace(session=fake_session)):
            with _swap_attr(chainlit_entry.cl, "user_session", fake_user_session):
                asyncio.run(chainlit_entry._on_message(_sample_message()))
    finally:
        runtime.register_discord_channel = original_register_discord_channel

    assert captured["thread_id"] == "thread-1"
    assert captured["channel_id"] == 777


def test_on_message_does_not_register_session_for_discord_client():
    runtime = get_runtime()
    app = EasierlitApp()
    client = EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread")
    runtime.bind(client=client, app=app)

    fake_session = SimpleNamespace(thread_id="thread-1", id="session-1", client_type="discord")
    fake_channel = SimpleNamespace(id=777)
    fake_user_session = SimpleNamespace(get=lambda key: fake_channel if key == "discord_channel" else None)

    calls = {"count": 0}
    original_register_session = runtime.register_session

    def fake_register_session(_thread_id: str, _session_id: str):
        calls["count"] += 1

    runtime.register_session = fake_register_session

    try:
        with _swap_attr(chainlit_entry.cl, "context", SimpleNamespace(session=fake_session)):
            with _swap_attr(chainlit_entry.cl, "user_session", fake_user_session):
                asyncio.run(chainlit_entry._on_message(_sample_message()))
    finally:
        runtime.register_session = original_register_session

    assert calls["count"] == 0


def test_on_message_registers_session_for_non_discord_client():
    runtime = get_runtime()
    app = EasierlitApp()
    client = EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread")
    runtime.bind(client=client, app=app)

    fake_session = SimpleNamespace(thread_id="thread-1", id="session-1", client_type="webapp")

    calls: list[tuple[str, str]] = []
    original_register_session = runtime.register_session

    def fake_register_session(thread_id: str, session_id: str):
        calls.append((thread_id, session_id))

    runtime.register_session = fake_register_session

    try:
        with _swap_attr(chainlit_entry.cl, "context", SimpleNamespace(session=fake_session)):
            asyncio.run(chainlit_entry._on_message(_sample_message()))
    finally:
        runtime.register_session = original_register_session

    assert calls == [("thread-1", "session-1")]


def test_on_message_enqueues_incoming_elements():
    runtime = get_runtime()
    app = EasierlitApp()
    client = EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread")
    runtime.bind(client=client, app=app)

    fake_session = SimpleNamespace(thread_id="thread-1", id="session-1", client_type="webapp")
    message = _sample_message()
    message.elements = [{"id": "el-1", "path": "/tmp/random.jpg"}]

    with _swap_attr(chainlit_entry.cl, "context", SimpleNamespace(session=fake_session)):
        asyncio.run(chainlit_entry._on_message(message))

    incoming = app.recv(timeout=1.0)
    assert incoming.elements == [{"id": "el-1", "path": "/tmp/random.jpg"}]


def test_start_and_stop_discord_bridge_lifecycle():
    runtime = get_runtime()
    runtime.bind(
        client=EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread"),
        app=EasierlitApp(),
        discord_token="discord-token",
    )

    started = {"count": 0}
    stopped = {"count": 0}

    class _FakeBridge:
        def __init__(self, *, runtime, bot_token):
            assert runtime is get_runtime()
            assert bot_token == "discord-token"

        async def start(self):
            started["count"] += 1

        async def stop(self):
            stopped["count"] += 1

    with _swap_attr(chainlit_entry, "EasierlitDiscordBridge", _FakeBridge):
        asyncio.run(chainlit_entry._start_discord_bridge_if_needed())
        assert started["count"] == 1

        asyncio.run(chainlit_entry._start_discord_bridge_if_needed())
        assert started["count"] == 2

        asyncio.run(chainlit_entry._stop_discord_bridge_if_running())
        assert stopped["count"] == 1


def test_on_app_startup_runs_local_storage_preflight_for_default_data_layer(
    tmp_path, monkeypatch
):
    runtime = get_runtime()
    runtime.unbind()
    runtime.bind(
        client=EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread"),
        app=EasierlitApp(),
    )

    monkeypatch.setenv("CHAINLIT_APP_ROOT", str(tmp_path))
    provider = LocalFileStorageClient(base_dir=tmp_path / "public" / "easierlit")
    fake_data_layer = SimpleNamespace(storage_provider=provider)
    calls = {"preflight": 0}

    monkeypatch.setattr(chainlit_entry, "_apply_runtime_configuration", lambda: None)
    monkeypatch.setattr(chainlit_entry, "get_data_layer", lambda: fake_data_layer)
    monkeypatch.setattr(chainlit_entry, "require_login", lambda: True)

    async def _noop_async():
        return None

    monkeypatch.setattr(chainlit_entry.RUNTIME, "start_dispatcher", _noop_async)
    monkeypatch.setattr(chainlit_entry, "_start_discord_bridge_if_needed", _noop_async)

    async def _fake_assert_local_storage_operational(storage_provider):
        del storage_provider
        calls["preflight"] += 1

    monkeypatch.setattr(
        chainlit_entry,
        "assert_local_storage_operational",
        _fake_assert_local_storage_operational,
    )

    chainlit_entry._DEFAULT_DATA_LAYER_REGISTERED = True
    asyncio.run(chainlit_entry._on_app_startup())

    assert calls["preflight"] == 1


def test_local_storage_route_serves_file_without_public_mount(tmp_path):
    runtime = get_runtime()
    persistence = EasierlitPersistenceConfig(
        enabled=True,
        sqlite_path=str(tmp_path / "route-test.db"),
        local_storage_dir=tmp_path / "outside",
    )
    storage_provider = _resolve_local_storage_provider(persistence)

    runtime.bind(
        client=EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread"),
        app=EasierlitApp(),
        persistence=persistence,
    )

    chainlit_entry._CONFIG_APPLIED = False
    chainlit_entry._apply_runtime_configuration()
    uploaded = asyncio.run(storage_provider.upload_file("user-1/image.png", b"payload"))

    with TestClient(chainlit_entry.chainlit_app) as client:
        response = client.get(uploaded["url"])

    assert uploaded["url"] == "/easierlit/local/user-1/image.png"
    assert response.status_code == 200
    assert response.content == b"payload"


def test_local_storage_route_missing_file_returns_404_not_spa_html(tmp_path):
    runtime = get_runtime()
    persistence = EasierlitPersistenceConfig(
        enabled=True,
        sqlite_path=str(tmp_path / "route-test.db"),
        local_storage_dir=tmp_path / "outside",
    )
    runtime.bind(
        client=EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread"),
        app=EasierlitApp(),
        persistence=persistence,
    )

    chainlit_entry._CONFIG_APPLIED = False
    chainlit_entry._LOCAL_STORAGE_PROVIDER = None
    chainlit_entry._LOCAL_STORAGE_ROUTE_REGISTERED = False
    chainlit_entry._apply_runtime_configuration()

    with TestClient(chainlit_entry.chainlit_app) as client:
        response = client.get("/easierlit/local/not-found/random.jpg")

    assert response.status_code == 404
    assert response.json()["detail"] == "File not found."


def test_local_storage_route_resolves_tilde_local_storage_dir(tmp_path, monkeypatch):
    runtime = get_runtime()
    monkeypatch.setenv("HOME", str(tmp_path))
    persistence = EasierlitPersistenceConfig(
        enabled=True,
        sqlite_path=str(tmp_path / "route-test.db"),
        local_storage_dir="~/fablit/workspace/images",
    )

    runtime.bind(
        client=EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread"),
        app=EasierlitApp(),
        persistence=persistence,
    )

    chainlit_entry._CONFIG_APPLIED = False
    chainlit_entry._LOCAL_STORAGE_PROVIDER = None
    chainlit_entry._apply_runtime_configuration()

    provider = _resolve_local_storage_provider(persistence)
    uploaded = asyncio.run(provider.upload_file("user-1/image.png", b"payload"))

    assert provider.base_dir == (tmp_path / "fablit" / "workspace" / "images").resolve()
    assert (provider.base_dir / "user-1" / "image.png").is_file()

    with TestClient(chainlit_entry.chainlit_app) as client:
        response = client.get(uploaded["url"])

    assert response.status_code == 200
    assert response.content == b"payload"


def test_local_storage_route_uses_runtime_persistence_provider_only(tmp_path, monkeypatch):
    runtime = get_runtime()
    persistence = EasierlitPersistenceConfig(
        enabled=True,
        sqlite_path=str(tmp_path / "route-test.db"),
        local_storage_dir=tmp_path / "persistence-images",
    )
    persistence_provider = _resolve_local_storage_provider(persistence)
    runtime.bind(
        client=EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread"),
        app=EasierlitApp(),
        persistence=persistence,
    )

    data_layer_provider = LocalFileStorageClient(base_dir=tmp_path / "actual-images")
    fake_data_layer = SimpleNamespace(storage_provider=data_layer_provider)
    monkeypatch.setattr(chainlit_entry, "get_data_layer", lambda: fake_data_layer)

    chainlit_entry._CONFIG_APPLIED = False
    chainlit_entry._apply_runtime_configuration()

    uploaded = asyncio.run(persistence_provider.upload_file("user-1/image.png", b"payload"))
    assert not (data_layer_provider.base_dir / "user-1" / "image.png").exists()

    with TestClient(chainlit_entry.chainlit_app) as client:
        response = client.get(uploaded["url"])

    assert response.status_code == 200
    assert response.content == b"payload"


def test_local_storage_route_uses_latest_runtime_persistence_provider_after_rebind(tmp_path):
    runtime = get_runtime()

    first_persistence = EasierlitPersistenceConfig(
        enabled=True,
        sqlite_path=str(tmp_path / "route-first.db"),
        local_storage_dir=tmp_path / "images-first",
    )
    first_provider = _resolve_local_storage_provider(first_persistence)
    runtime.bind(
        client=EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread"),
        app=EasierlitApp(),
        persistence=first_persistence,
    )
    chainlit_entry._CONFIG_APPLIED = False
    chainlit_entry._LOCAL_STORAGE_PROVIDER = None
    chainlit_entry._LOCAL_STORAGE_ROUTE_REGISTERED = False
    chainlit_entry._apply_runtime_configuration()
    first_uploaded = asyncio.run(first_provider.upload_file("user-1/first.png", b"first"))

    with TestClient(chainlit_entry.chainlit_app) as client:
        first_response = client.get(first_uploaded["url"])
    assert first_response.status_code == 200
    assert first_response.content == b"first"

    second_persistence = EasierlitPersistenceConfig(
        enabled=True,
        sqlite_path=str(tmp_path / "route-second.db"),
        local_storage_dir=tmp_path / "images-second",
    )
    second_provider = _resolve_local_storage_provider(second_persistence)
    runtime.unbind()
    runtime.bind(
        client=EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread"),
        app=EasierlitApp(),
        persistence=second_persistence,
    )
    chainlit_entry._CONFIG_APPLIED = False
    # Intentionally keep route registration and previous provider state.
    second_uploaded = asyncio.run(second_provider.upload_file("user-2/second.png", b"second"))

    with TestClient(chainlit_entry.chainlit_app) as client:
        second_response = client.get(second_uploaded["url"])
    assert second_response.status_code == 200
    assert second_response.content == b"second"


def test_default_sqlite_data_layer_get_thread_refreshes_element_url_from_object_key(
    tmp_path, monkeypatch
):
    runtime = get_runtime()
    runtime.unbind()
    config.code.data_layer = None

    previous_database_url = os.environ.get("DATABASE_URL")
    previous_literal_api_key = os.environ.get("LITERAL_API_KEY")
    os.environ.pop("DATABASE_URL", None)
    os.environ.pop("LITERAL_API_KEY", None)

    chainlit_entry._CONFIG_APPLIED = False
    chainlit_entry._LOCAL_STORAGE_PROVIDER = None

    class _FakeSQLAlchemyDataLayer:
        def __init__(self, conninfo: str, storage_provider=None):
            self._conninfo = conninfo
            self.storage_provider = storage_provider

        async def get_thread(self, thread_id: str):
            return {
                "id": thread_id,
                "elements": [
                    {
                        "id": "el-1",
                        "objectKey": "user-1/image.png",
                        "url": "/stale/url",
                    },
                    {
                        "id": "el-2",
                        "url": "https://example.com/keep.png",
                    },
                ],
            }

    class _FakeStorageProvider:
        async def get_read_url(self, object_key: str) -> str:
            return f"/easierlit/local/{object_key}"

    monkeypatch.setattr(chainlit_sql_alchemy, "SQLAlchemyDataLayer", _FakeSQLAlchemyDataLayer)

    db_path = tmp_path / "refresh-thread-url.db"
    runtime.bind(
        client=EasierlitClient(run_funcs=[lambda _app: None], worker_mode="thread"),
        app=EasierlitApp(),
        persistence=EasierlitPersistenceConfig(enabled=True, sqlite_path=str(db_path)),
    )

    try:
        chainlit_entry._apply_runtime_configuration()
        assert config.code.data_layer is not None

        data_layer = config.code.data_layer()
        data_layer.storage_provider = _FakeStorageProvider()
        thread = asyncio.run(data_layer.get_thread("thread-1"))

        assert thread["elements"][0]["url"] == "/easierlit/local/user-1/image.png"
        assert thread["elements"][1]["url"] == "https://example.com/keep.png"
    finally:
        runtime.unbind()
        config.code.data_layer = None
        chainlit_entry._CONFIG_APPLIED = False
        chainlit_entry._LOCAL_STORAGE_PROVIDER = None

        if previous_database_url is None:
            os.environ.pop("DATABASE_URL", None)
        else:
            os.environ["DATABASE_URL"] = previous_database_url

        if previous_literal_api_key is None:
            os.environ.pop("LITERAL_API_KEY", None)
        else:
            os.environ["LITERAL_API_KEY"] = previous_literal_api_key
